#include <cstdlib>

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include <vector>
#include <cmath>

#include <stdio.h>

// gsl
#include <gsl/gsl_fit.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_cdf.h>

using namespace std;

#include "constants.h"

double missing_code= -999.0;  // code for missing T and P values

string samplenames[max_sample];
double outcome[max_sample];

typedef vector<vector<double> > matrix_t;
matrix_t cov_data;
void sizeMatrix(matrix_t &, int,int);

int coord[max_coord];
double tvalue_vec[max_coord];
double pvalue_vec[max_coord];

string pheno_file, regression_label;
int n_sample, n_coord, n_cov;

void gsl_multifit_linear_wrapper (double* outcome, double* cpgdata, \
                                  matrix_t& covariates, int n_covariates, int n_sample, \
                                  double* t_value, double* p_value);


void read_pheno(const char * filename);
void read_coord(const char * filename);
void cov_matrix_dimensions(const char * filename);


int main(int argc, char *argv[])
{

	#ifdef __WIN__
	  // Windows stuff
	  _setmaxstdio(2048);
	#endif

    string pheno_file = argv[2];        // data file which assumes structure: sample ID, outcome variable, covariates
    int ichrom = atoi(argv[5]);  // chromosome number
    string regression_label  = argv[4];  // a string that characterizes the analysis and is appended to the output file
    string covdir     = argv[1]; // directory where the coverage files are

	string workdir = argv[3];

    cout << " The chromosome is " << ichrom << "\n";
    cout << " The regression label is " << regression_label << "\n";

    int i, j, iter;
    string STRING, subst, filna;
	size_t pos, oldpos;
    stringstream ss;
	const char *ptemp;

	// read the phenotype file
    ss << pheno_file;  // use abosolute path by chen
    filna = ss.str();
    ss.str(string());
    ptemp= filna.c_str();
    cov_matrix_dimensions(ptemp);  // n_sample and n_cov are determined
    if (n_cov > 0) {
       sizeMatrix(cov_data, n_sample,n_cov);
       read_pheno(ptemp);
    }
    else {
	   sizeMatrix(cov_data, n_sample, 1);
       read_pheno(ptemp);
	//	 cout << " Exit program because there are no covariates in the file " << ptemp << "\n";exit (1);
	}

    ifstream infils[n_sample];
    double cpgdata[n_sample];

   // read first sample to determine number of coordinates
   int sample_nr = 0;
   ss << covdir << "chr" << ichrom << "/" << samplenames[sample_nr] << ".csv";
   filna.clear();
   filna = ss.str();
   ss.str(string());
   ptemp= filna.c_str();
   (infils[sample_nr]).open(ptemp);
   if (!(infils[sample_nr]).is_open()) {cout << "file named " << ptemp << "  is not open \n";exit (1);}
   read_coord(ptemp);
   cout << " The number of coordinates is: "  << n_coord << "\n";
   infils[sample_nr].close();

    // Open coverage files
  	for(i=0;i<n_sample;i++) {
       ss << covdir << "chr" << ichrom << "/" << samplenames[i] << ".csv";
       filna.clear();
       filna = ss.str();
       ss.str(string());
       ptemp= filna.c_str();
       (infils[i]).open(ptemp);
       if (!(infils[i]).is_open()) {cout << "file named " << ptemp << "  is not open \n"; exit (1);}
        }
    for(i=0;i<n_sample;i++) getline(infils[i],STRING); // Skipping the headers


     double t_value,p_value;
    // Here starts the loop on coordinates;
    for(iter=0;iter<n_coord;iter++) {


       for(i=0;i<n_sample;i++) {
  	      getline(infils[i],STRING);
          oldpos = STRING.find(',');
          pos=STRING.find(',',oldpos+1);
          subst = STRING.substr(oldpos+1,pos-oldpos-1);
          cpgdata[i] = atof(subst.c_str());
        }

     // multiple linear regression
       double* covariates = NULL;
       int n_covariates = 0;
       gsl_multifit_linear_wrapper (outcome, cpgdata, \
                                     cov_data, n_cov, n_sample, \
                                     &t_value, &p_value);

        tvalue_vec[iter] = t_value;
        pvalue_vec[iter] = p_value;

    }  // end   for(iter=0;iter<n_coord;iter++)
    for(i=0;i<n_sample;i++) (infils[i]).close();

    // Write results
    ss << workdir << "association_" << regression_label << "_chr" << ichrom << ".csv";
    filna = ss.str();
    ss.str(string());
    cout << " The name of the output file is \n";
    cout << filna << "\n";
    ptemp= filna.c_str();

    ofstream outfile;
    outfile.open(ptemp);
    outfile.setf(ios::scientific);
    outfile.precision(15);
    outfile << "coord,t_value,p_value" << "\n"; // Creating header
    for(i=0;i<n_coord;i++) outfile << coord[i] << "," << tvalue_vec[i] << "," << pvalue_vec[i] << "\n";
    outfile.close();

    return EXIT_SUCCESS;
}

//---------functions----------------


void gsl_multifit_linear_wrapper (double* outcome, double* cpgdata, \
                                  matrix_t& covariates, int n_covariates, int n_sample, \
                                  double* t_value, double* p_value) {
    gsl_matrix *X, *cov;
    gsl_vector *y, *c;

    double chisq;
    int n = n_sample;
    int p = 2 + n_covariates; // intercept + outcome + covariates
    // allocate the memory
    X = gsl_matrix_alloc (n, p);
    y = gsl_vector_alloc (n);

    c = gsl_vector_alloc (p);
    cov = gsl_matrix_alloc (p, p);
    // setup the data
    for (int i=0; i<n; i++){
        // fill X
        gsl_matrix_set(X, i, 0, 1.0);
        gsl_matrix_set(X, i, 1, outcome[i]);
        for (int j=0; j<n_covariates; j++){
            gsl_matrix_set(X, i, 2+j, covariates[i][j]);
        }
        // fill y
        gsl_vector_set(y, i, cpgdata[i]);

    }

    // invoke the gls function for weighted multiple regression
    {
        gsl_multifit_linear_workspace * work \
               = gsl_multifit_linear_alloc (n, p);


        gsl_multifit_linear (X, y, c, cov, \
                                   &chisq, work);

        gsl_multifit_linear_free (work);
    }

   // calculate the p value for variable x
   if (chisq == 0)
     {*t_value=missing_code;
      *p_value=missing_code; } else {
         *t_value = gsl_vector_get(c,1)/sqrt(gsl_matrix_get(cov,(1),(1)));
         int df = n - p;
         *p_value = 2*gsl_cdf_tdist_P(-abs(*t_value), df);
     }

    // free the memory
    gsl_matrix_free (X);
    gsl_vector_free (y);
    gsl_vector_free (c);
    gsl_matrix_free (cov);

}

void cov_matrix_dimensions(const char * filename) {

    string STRING, subst, subst2;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
	int i=0, j;

	infile.open (filename);
    if (!(infile).is_open()) {cout << "The file " << filename << "  is not open \n"; exit (1);}
       else cout << " The name of the phenotype file is " <<  filename  << "\n";

	getline(infile,STRING); // Determine number of covariates
	oldpos = 0;
	pos=STRING.find(',');
    j=0;
    while (pos!=string::npos) {
       oldpos = pos+1;
       pos=STRING.find(',',oldpos);
       subst = STRING.substr(oldpos,pos-oldpos);
       j++;
    }
    n_cov = j - 1;
    cout << " Number of covariates is " << n_cov << "\n";

    while(!infile.eof()) {
	        getline(infile,STRING);
	        if (STRING=="") break;
            i++;
    }

    n_sample = i ;
    cout << " Number of samples is " << n_sample << "\n";
    infile.close();
}

void read_pheno(const char * filename) {

    string STRING, subst, subst2;
	ifstream infile;
	size_t pos, oldpos, pos1, pos2;
	int i=0, j;

	infile.open (filename);
    if (!(infile).is_open()) {cout << "The file " << filename << "  is not open \n"; exit (1);}
	getline(infile,STRING); // Skip header

    while(!infile.eof()) {
	   getline(infile,STRING);
	   if (STRING=="") break;
	   oldpos = 0;
	   pos=STRING.find(',');
	   if (pos==string::npos) break;
       subst = STRING.substr(oldpos,pos-oldpos);
       samplenames[i] = subst;   // creating the names of all coverage files

       j=-1;
	    while (pos!=string::npos) {
           oldpos = pos+1;
           pos=STRING.find(',',oldpos);
           subst = STRING.substr(oldpos,pos-oldpos);
           if (j==-1) outcome[i] = atof(subst.c_str());
             else cov_data[i][j]  = atof(subst.c_str());
           j++;
            }

       i++;
    }

    infile.close();
}


void sizeMatrix(matrix_t & m , int r, int c)
{
  m.clear();
  m.resize(r);
  for (int i=0; i<r; i++)
    m[i].resize(c,0);
}



void read_coord(const char * filename) {

    string STRING, subst;
	ifstream infile;
	size_t pos, oldpos;


	infile.open(filename);

	getline(infile,STRING); // Skipping the header

	n_coord = 0;
    while(!infile.eof()) {
        getline(infile,STRING);
        if (STRING=="") break;
        oldpos = 0;
	    pos=STRING.find(',');
        subst = STRING.substr(oldpos,pos-oldpos);
        coord[n_coord] = atoi(subst.c_str());
        n_coord++;
      }

    infile.close();

}
